/*
 * pt_deny_attach.c
 *
 * Author: Landon J. Fuller <landonf@opendarwin.org>
 * This software is placed in the public domain
 * 
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

#include <mach/mach_types.h>
#include <sys/systm.h>
#include <sys/kernel.h>

#define PT_DENY_ATTACH  31

/*
 * ptrace argument structure
 * XXX This must stay in sync with the real structure in xnu/bsd/sys/sysproto.h
 */
#if BYTE_ORDER == BIG_ENDIAN
struct ptrace_args {
	char req_l_[(sizeof(uint64_t) <= sizeof(int) ? 0 : sizeof(uint64_t) - sizeof(int))];
	int req;
	char req_r_[0];
 
	char pid_l_[(sizeof(uint64_t) <= sizeof(pid_t) ? 0 : sizeof(uint64_t) - sizeof(pid_t))];
	pid_t pid;
	char pid_r_[0];
 
	char addr_l_[(sizeof(uint64_t) <= sizeof(user_addr_t) ? 0 : sizeof(uint64_t) - sizeof(user_addr_t))];
	user_addr_t addr;
	char addr_r_[0];
 
	char data_l_[(sizeof(uint64_t) <= sizeof(int) ? 0 : sizeof(uint64_t) - sizeof(int))];
	int data;
	char data_r_[0];
};

#else /* Little Endian */

struct ptrace_args {
	char req_l_[0];
	int req;
	char req_r_[(sizeof(uint64_t) <= sizeof(int) ? 0 : sizeof(uint64_t) - sizeof(int))];
 
	char pid_l_[0];
	pid_t pid;
	char pid_r_[(sizeof(uint64_t) <= sizeof(pid_t) ? 0 : sizeof(uint64_t) - sizeof(pid_t))];

	char addr_l_[0];
	user_addr_t addr;
	char addr_r_[(sizeof(uint64_t) <= sizeof(user_addr_t) ? 0 : sizeof(uint64_t) - sizeof(user_addr_t))];

	char data_l_[0];
	int data;
	char data_r_[(sizeof(uint64_t) <= sizeof(int) ? 0 : sizeof(uint64_t) - sizeof(int))];
};

#endif /* BYTE_ORDER */


int (*real_ptrace) (struct proc *, struct ptrace_args *, register_t *);

int our_ptrace (struct proc *p, struct ptrace_args *uap, register_t *retval)
{
	int ret;
	
	if (uap->req == PT_DENY_ATTACH)
		return (0);
	else
		return real_ptrace(p, uap, retval);
		
	return ret;
}

kern_return_t pt_deny_attach_start (kmod_info_t * ki, void * d) {
	/* If ptrace is already patched, unpatch it */
	temp_unpatch_ptrace();
	
	/* Patch ptrace() */
	real_ptrace = (void *) temp_patch_ptrace((uintptr_t) our_ptrace);
	
	printf("Disallowing ptrace(PT_DENY_ATTACH, ...)\n");

	return KERN_SUCCESS;
}


kern_return_t pt_deny_attach_stop (kmod_info_t * ki, void * d) {
	/* Unpatch ptrace() */
	temp_unpatch_ptrace();
	
	printf("Allowing ptrace(PT_DENY_ATTACH, ...)\n");

	return KERN_SUCCESS;
}
